@@ -1,6 +1,6 @@
 
 /*
- * Copyright 2013--2015 Jan Pazdziora
+ * Copyright 2013--2016 Jan Pazdziora
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -47,11 +47,13 @@
 #define DBUS_SSSD_USERS_ID "name";
 #endif
 
-static const int LOOKUP_IDENTITY_OUTPUT_NONE = 0;
+static const int LOOKUP_IDENTITY_OUTPUT_DEFAULT = 0;
+
+static const int LOOKUP_IDENTITY_OUTPUT_NONE = 128;
 static const int LOOKUP_IDENTITY_OUTPUT_NOTES = 1;
 static const int LOOKUP_IDENTITY_OUTPUT_ENV = 2;
-static const int LOOKUP_IDENTITY_OUTPUT_ALL = 3;
-static const int LOOKUP_IDENTITY_OUTPUT_DEFAULT = 4;
+static const int LOOKUP_IDENTITY_OUTPUT_HEADERS = 4;
+static const int LOOKUP_IDENTITY_OUTPUT_HEADERS_BASE64 = 8;
 
 static char * LOOKUP_IDENTITY_OUTPUT_GECOS = "REMOTE_USER_GECOS";
 
@@ -67,7 +69,7 @@ typedef struct lookup_identity_config {
 	apr_hash_t * output_user_attr_sep;
 	apr_hash_t * output_user_attr_iter;
 	int dbus_timeout;
-	signed char lookup_by_certificate;
+	int lookup_by_certificate;
 #endif
 } lookup_identity_config;
 
@@ -110,10 +112,10 @@ static int lookup_user_by_certificate(request_rec * r) {
 		goto fail;
 	}
 
-        int timeout = DBUS_SSSD_TIMEOUT;
-        if (cfg->dbus_timeout > 0) {
-                timeout = cfg->dbus_timeout;
-        }
+	int timeout = DBUS_SSSD_TIMEOUT;
+	if (cfg->dbus_timeout > 0) {
+		timeout = cfg->dbus_timeout;
+	}
 
 	reply = dbus_connection_send_with_reply_and_block(connection,
 		message, timeout, &error);
@@ -303,14 +305,25 @@ static DBusMessage * lookup_identity_dbus_message(request_rec * r, DBusConnectio
 }
 #endif
 
-static void lookup_identity_output_iter_to(request_rec * r, apr_table_t * t, const char * key, const apr_array_header_t * values) {
+static apr_array_header_t * base64_encode_array(apr_pool_t * p, const apr_array_header_t * values) {
+	if (! values)
+		return NULL;
+	apr_array_header_t * base64_values = apr_array_make(p, values->nelts, sizeof(char *));
+	for (int i = 0; i < values->nelts; i++) {
+		*(char **)apr_array_push(base64_values) = ap_pbase64encode(p, ((char **)values->elts)[i]);
+	}
+	ap_assert(values->nelts == base64_values->nelts);
+	return base64_values;
+}
+
+static void lookup_identity_output_iter_to(request_rec * r, apr_table_t * t, const char * key, const char * sep, const apr_array_header_t * values) {
 	int append = 0;
 	if (key[0] == '+') {
 		key++;
 		append = 1;
 	}
 	long start = 0;
-	const char * key_n = apr_pstrcat(r->pool, key, "_N", NULL);
+	const char * key_n = apr_pstrcat(r->pool, key, sep, "N", NULL);
 	if (append) {
 		const char * start_index = apr_table_get(t, key_n);
 		if (start_index) {
@@ -318,16 +331,21 @@ static void lookup_identity_output_iter_to(request_rec * r, apr_table_t * t, con
 		}
 	}
 	for (int i = 0; values && i < values->nelts; i++) {
-		apr_table_setn(t, apr_psprintf(r->pool, "%s_%ld", key, ++start), apr_pstrdup(r->pool, ((char **)values->elts)[i]));
+		apr_table_setn(t, apr_psprintf(r->pool, "%s%s%ld", key, sep, ++start), apr_pstrdup(r->pool, ((char **)values->elts)[i]));
 	}
 	apr_table_setn(t, key_n, apr_psprintf(r->pool, "%ld", start));
 }
 static void lookup_identity_output_iter(request_rec * r, int the_output, const char * key, const apr_array_header_t * values) {
 	if (the_output & LOOKUP_IDENTITY_OUTPUT_NOTES) {
-		lookup_identity_output_iter_to(r, r->notes, key, values);
+		lookup_identity_output_iter_to(r, r->notes, key, "_", values);
 	}
 	if (the_output & LOOKUP_IDENTITY_OUTPUT_ENV) {
-		lookup_identity_output_iter_to(r, r->subprocess_env, key, values);
+		lookup_identity_output_iter_to(r, r->subprocess_env, key, "_", values);
+	}
+	if (the_output & LOOKUP_IDENTITY_OUTPUT_HEADERS_BASE64) {
+		lookup_identity_output_iter_to(r, r->headers_in, key, "-", base64_encode_array(r->pool, values));
+	} else if (the_output & LOOKUP_IDENTITY_OUTPUT_HEADERS) {
+		lookup_identity_output_iter_to(r, r->headers_in, key, "-", values);
 	}
 }
 
@@ -365,6 +383,11 @@ static void lookup_identity_output_data(request_rec * r, int the_output, const c
 	if (the_output & LOOKUP_IDENTITY_OUTPUT_ENV) {
 		lookup_identity_output_data_to(r, r->subprocess_env, key, values, sep);
 	}
+	if (the_output & LOOKUP_IDENTITY_OUTPUT_HEADERS_BASE64) {
+		lookup_identity_output_data_to(r, r->headers_in, key, base64_encode_array(r->pool, values), sep);
+	} else if (the_output & LOOKUP_IDENTITY_OUTPUT_HEADERS) {
+		lookup_identity_output_data_to(r, r->headers_in, key, values, sep);
+	}
 }
 
 static void * merge_dir_conf(apr_pool_t * pool, void * base_void, void * add_void);
@@ -393,8 +416,8 @@ static int lookup_identity_hook(request_rec * r) {
 
 	int the_output = the_config->output;
 	if (the_output == LOOKUP_IDENTITY_OUTPUT_DEFAULT) {
-		the_output = LOOKUP_IDENTITY_OUTPUT_ALL;
-	} else if (!(the_output & LOOKUP_IDENTITY_OUTPUT_ALL)) {
+		the_output = LOOKUP_IDENTITY_OUTPUT_NOTES | LOOKUP_IDENTITY_OUTPUT_ENV;
+	} else if (the_output & LOOKUP_IDENTITY_OUTPUT_NONE) {
 		return DECLINED;
 	}
 
@@ -570,11 +593,17 @@ static const char * set_output(cmd_parms * cmd, void * conf_void, const char * a
 		if (!strcasecmp(arg, "none")) {
 			cfg->output = LOOKUP_IDENTITY_OUTPUT_NONE;
 		} else if (!strcasecmp(arg, "all")) {
-			cfg->output = LOOKUP_IDENTITY_OUTPUT_ALL;
+			cfg->output |= LOOKUP_IDENTITY_OUTPUT_ENV | LOOKUP_IDENTITY_OUTPUT_NOTES;
+			ap_log_error(APLOG_MARK, APLOG_ERR, 0, cmd->server,
+				"LookupOutput all is deprecated, use none, env, notes, or headers");
 		} else if (!strcasecmp(arg, "env")) {
-			cfg->output = LOOKUP_IDENTITY_OUTPUT_ENV;
+			cfg->output |= LOOKUP_IDENTITY_OUTPUT_ENV;
 		} else if (!strcasecmp(arg, "notes")) {
-			cfg->output = LOOKUP_IDENTITY_OUTPUT_NOTES;
+			cfg->output |= LOOKUP_IDENTITY_OUTPUT_NOTES;
+		} else if (!strcasecmp(arg, "headers-base64")) {
+			cfg->output |= LOOKUP_IDENTITY_OUTPUT_HEADERS_BASE64;
+		} else if (!strcasecmp(arg, "headers")) {
+			cfg->output |= LOOKUP_IDENTITY_OUTPUT_HEADERS;
 		}
 	}
 	return NULL;
@@ -701,7 +730,7 @@ static void * merge_dir_conf(apr_pool_t * pool, void * base_void, void * add_voi
 }
 
 static const command_rec directives[] = {
-	AP_INIT_TAKE1("LookupOutput", set_output, NULL, RSRC_CONF | ACCESS_CONF, "Specify where the lookup results should be stored (note or variable)"),
+	AP_INIT_TAKE1("LookupOutput", set_output, NULL, RSRC_CONF | ACCESS_CONF, "Specify where the lookup results should be stored (notes, variables, headers)"),
 	AP_INIT_TAKE1("LookupUserGECOS", ap_set_string_slot, (void*)APR_OFFSETOF(lookup_identity_config, output_gecos), RSRC_CONF | ACCESS_CONF, "Name of the note/variable for the GECOS information"),
 #ifndef NO_USER_ATTR
 	AP_INIT_TAKE12("LookupUserGroups", set_output_groups, NULL, RSRC_CONF | ACCESS_CONF, "Name of the note/variable for the group information"),
@@ -709,17 +738,21 @@ static const command_rec directives[] = {
 	AP_INIT_TAKE23("LookupUserAttr", set_user_attr, NULL, RSRC_CONF | ACCESS_CONF, "Additional user attribute (attr, note/variable name, separator)"),
 	AP_INIT_TAKE2("LookupUserAttrIter", set_user_attr_iter, NULL, RSRC_CONF | ACCESS_CONF, "Additional user attributes (attr, note/variable name)"),
 	AP_INIT_TAKE1("LookupDbusTimeout", ap_set_int_slot, (void*)APR_OFFSETOF(lookup_identity_config, dbus_timeout), RSRC_CONF | ACCESS_CONF, "Timeout for sssd dbus calls (in ms)"),
-	AP_INIT_FLAG("LookupUserByCertificate", ap_set_flag_slot_char, (void*)APR_OFFSETOF(lookup_identity_config, lookup_by_certificate), RSRC_CONF | ACCESS_CONF, "Use org.freedesktop.sssd.infopipe.Users.FindByCertificate to lookup user identity"),
+	AP_INIT_FLAG("LookupUserByCertificate", ap_set_flag_slot, (void*)APR_OFFSETOF(lookup_identity_config, lookup_by_certificate), RSRC_CONF | ACCESS_CONF, "Use org.freedesktop.sssd.infopipe.Users.FindByCertificate to lookup user identity"),
 #endif
 	{ NULL }
 };
 
 static void register_hooks(apr_pool_t * pool) {
 #ifndef NO_USER_ATTR
-        static const char * const access_succ[] = {"mod_authz_core.c", NULL};
-        static const char * const access_pred[] = {"mod_ssl.c", NULL};
-        ap_hook_check_access(lookup_user_by_certificate, access_pred, access_succ, APR_HOOK_MIDDLE,
-                                        AP_AUTH_INTERNAL_PER_CONF);
+	static const char * const access_succ[] = {"mod_authz_core.c", NULL};
+	static const char * const access_pred[] = {"mod_ssl.c", "mod_nss.c", NULL};
+#ifdef AP_AUTH_INTERNAL_PER_CONF
+	ap_hook_check_access(lookup_user_by_certificate, access_pred, access_succ, APR_HOOK_MIDDLE,
+					AP_AUTH_INTERNAL_PER_CONF);
+#else
+	ap_hook_access_checker(lookup_user_by_certificate, access_pred, access_succ, APR_HOOK_MIDDLE);
+#endif
 #endif
 
 	static const char * const fixup_succ[] = {"mod_headers.c", NULL};
